def get_percent_change(new_value, old_value):
    if not old_value and not new_value:
        return 0
    if not old_value:
        sign = 1 if new_value > 0 else -1
        return sign * float('inf')
    change_difference = new_value - old_value
    change_ratio = change_difference / float(old_value)
    return 100 * change_ratio


def define_normalize(xs, ys):
    x_min, x_max = min(xs), max(xs)
    y_min, y_max = min(ys), max(ys)

    x_width = (x_max - x_min) or 1
    y_width = (y_max - y_min) or 1

    def normalize(x):
        unit_x = (x - x_min) / float(x_width)
        return (unit_x * y_width) + y_min

    return normalize


def divide_safely(numerator, denominator, default):
    if not denominator:
        if isinstance(default, Exception):
            raise default
        return default
    return numerator / float(denominator)


def round_number(x):
    return int(round(abs(x)))



from copy import copy


def flip_geometry_coordinates(geometries):
    flipped_geometries = []
    for geometry in geometries:
        if hasattr(geometry, 'geoms'):
            flipped_geometry = geometry.__class__(
                flip_geometry_coordinates(geometry.geoms))
        else:
            flipped_geometry = copy(geometry)
            flipped_geometry.coords = [flip_xy(xyz) for xyz in geometry.coords]
        flipped_geometries.append(flipped_geometry)
    return flipped_geometries


def flip_xy(xyz):
    'Flip x and y coordinates whether or not there is a z-coordinate'
    xyz = list(xyz)  # Preserve original
    xyz[0], xyz[1] = xyz[1], xyz[0]
    return xyz



import random
import string


def make_random_string(length, with_punctuation=False, with_spaces=False):
    alphabet = string.digits + string.letters
    if with_punctuation:
        alphabet += string.punctuation
    if with_spaces:
        alphabet += ' '
    return ''.join(random.choice(alphabet) for x in range(length))



import shlex
import subprocess
from six import string_types

from .exceptions import InvisibleRoadsError


def run_command(command, exception_by_error=None):
    if not isinstance(command, string_types):
        command = ' '.join(command)
    command = command.split(';', 1)[0]
    return run_raw_command(command, exception_by_error)


def run_raw_command(command, exception_by_error=None):
    if isinstance(command, string_types):
        command = shlex.split(command)
    try:
        output = subprocess.check_output(command, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        o = e.output
        for error_text, exception in (exception_by_error or {}).items():
            if error_text in o:
                raise exception
        else:
            raise InvisibleRoadsError(o)
    except OSError as e:
        raise InvisibleRoadsError(e.strerror)
    return output.strip()



import functools
import re
import shlex
import sys
from importlib import import_module
from six.moves.configparser import NoSectionError, RawConfigParser

from .iterable import merge_dictionaries


class RawCaseSensitiveConfigParser(RawConfigParser):
    optionxform = str


def load_settings(configuration_path, section_name):
    configuration = RawCaseSensitiveConfigParser()
    configuration.read(configuration_path)
    try:
        d = dict(configuration.items(section_name))
    except NoSectionError:
        d = {}
    return d


def parse_settings(settings, prefix, parse_setting=None):
    d = {}
    prefix_pattern = re.compile('^' + prefix.replace('.', r'\.'))
    if not parse_setting:
        parse_setting = parse_raw_setting
    for k, v in settings.items():
        if not k.startswith(prefix):
            continue
        d = merge_dictionaries(d, parse_setting(prefix_pattern.sub('', k), v))
    return d


def parse_raw_setting(k, v):
    return {k: v}


def resolve_attribute(attribute_spec):
    # Modified from pkg_resources.EntryPoint.resolve()
    if not attribute_spec or not hasattr(attribute_spec, 'split'):
        return attribute_spec
    module_url, attributes_string = attribute_spec.split(':')
    module = import_module(module_url)
    try:
        attribute = functools.reduce(
            getattr, attributes_string.split('.'), module)
    except AttributeError:
        raise ImportError('could not resolve attribute (%s)' % module_url)
    return attribute


def split_arguments(command_string):
    # Strip line breaks: \ for POSIX, ^ for Windows
    lines = []
    for line in command_string.splitlines():
        lines.append(line.rstrip(' \\^'))
    string = ' '.join(lines)
    # Split terms
    try:
        xs = shlex.split(string)
    except UnicodeEncodeError:
        xs = shlex.split(string.encode('utf-8'))
    return [x.strip() for x in xs]


def unicode_safely(x):
    # http://stackoverflow.com/a/23085282/192092
    if not hasattr(x, 'decode'):
        return x
    return x.decode(sys.getfilesystemencoding())



from collections import Callable, MutableSet, OrderedDict
from copy import deepcopy


class OrderedDefaultDict(OrderedDict):
    # http://stackoverflow.com/a/6190500/192092

    def __init__(self, default_factory=None, *a, **kw):
        if (default_factory is not None and
                not isinstance(default_factory, Callable)):
            raise TypeError('first argument must be callable')
        OrderedDict.__init__(self, *a, **kw)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = self.default_factory,
        return type(self), args, None, None, self.items()

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return type(self)(self.default_factory, self)

    def __deepcopy__(self, memo):
        return type(self)(self.default_factory, deepcopy(self.items()))

    def __repr__(self):
        return 'OrderedDefaultDict(%s, %s)' % (
            self.default_factory, OrderedDict.__repr__(self))


class OrderedSet(MutableSet):
    'Written by Raymond Hettinger'
    # http://code.activestate.com/recipes/576694/

    def __init__(self, iterable=None):
        self.end = end = []
        end += [None, end, end]         # sentinel node for doubly linked list
        self.map = {}                   # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:
            key, prev, next = self.map.pop(key)
            prev[2] = next
            next[1] = prev

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)


def flatten_lists(list_of_lists):
    # http://stackoverflow.com/a/952952/192092
    return [item for sublist in list_of_lists for item in sublist]


def flatten_dictionaries(dictionary_of_dictionaries):
    'Combined nested dictionary into a simple dictionary'
    d = OrderedDict()
    for outer_k, outer_v in dictionary_of_dictionaries.items():
        if not hasattr(outer_v, 'items'):
            d[outer_k] = outer_v
            continue
        for inner_k, inner_v in flatten_dictionaries(outer_v).items():
            d['%s.%s' % (outer_k, inner_k)] = inner_v
    return d


def merge_dictionaries(*dictionaries):
    'Combine multiple nested dictionaries and overwrite duplicate keys'
    d = OrderedDict()
    for outer_d in dictionaries:
        for outer_k, outer_v in outer_d.items():
            d1 = d.get(outer_k, {})
            d2 = outer_v
            if hasattr(d1, 'items') and hasattr(d2, 'items'):
                d[outer_k] = merge_dictionaries(d1, d2)
            else:
                d[outer_k] = d2
    return d


def get_lists_from_tuples(xs):
    'Convert tuples to lists'
    # http://stackoverflow.com/a/1014669
    if isinstance(xs, (list, tuple)):
        return list(map(get_lists_from_tuples, xs))
    return xs


def set_default(settings, key, default, parse=None):
    'Set key with default if it does not exist or parse value if it exists'
    if key not in settings:
        value = default
    elif parse:
        value = parse(settings[key])
    settings[key] = value
    return value



import re


WHITESPACE_PATTERN = re.compile(r'\s+', re.MULTILINE)
PUNCTUATION_PATTERN = re.compile(r'[^a-zA-Z\s]+')


def parse_words(x):
    return x.replace(',', ' ').split()


def strip_whitespace(string):
    return WHITESPACE_PATTERN.sub('', string)


def compact_whitespace(string):
    return WHITESPACE_PATTERN.sub(' ', string).strip()


def remove_punctuation(string):
    return compact_whitespace(PUNCTUATION_PATTERN.sub(' ', string))



# http://stackoverflow.com/a/1119769/192092
PRONOUNCEABLE_ALPHABET = '23456789abcdefghijkmnpqrstuvwxyz'


def encode_number(non_negative_integer, alphabet=PRONOUNCEABLE_ALPHABET):
    if non_negative_integer < 0:
        raise ValueError
    if non_negative_integer == 0:
        return alphabet[0]
    characters = []
    base = len(alphabet)
    while non_negative_integer:
        remainder = non_negative_integer % base
        non_negative_integer = non_negative_integer // base
        characters.append(alphabet[remainder])
    characters.reverse()
    return ''.join(characters)


def decode_number(string, alphabet=PRONOUNCEABLE_ALPHABET):
    base = len(alphabet)
    string_length = len(string)
    number = 0
    try:
        for character_number, character in enumerate(string, 1):
            power = string_length - character_number
            number += alphabet.index(character) * (base ** power)
    except ValueError:
        number = -1
    return number
